/*This is a subject of a project entitled as "Hapmap Accession Distinguishment through constructing a Hierarchical Tree"
**In this sub-project, the genotype matrix markers will be extracted and filtering.
  The relationship of the files are following:
 inputting file: BCF_Text file, which are output from .bcf file by the perl script.
 output files:   1. Genotype Matrix File, 2. Marker_Atlas, 3. Accession_Atlas.
 Written By Wenchao Zhang 01/26/2017
 Copy Right: Noble Foundation
*/
#include <vector>
#include <algorithm>
using namespace std;
#include <iostream>
#include <sstream>
#include <fstream>
#include <string>

int parse_cmd_line(int argc, char* argv[], string & Filename_BCF_Text, string & Filename_Genotype_Matrix, string & Filename_Accession_Atlas, string & Filename_Marker_Atlas, int & Shif_Col, int & Min_AF, int & Max_Unknow, int & Max_Hetero, int & Max_Length_Dif_Allele);
int GMEF(string Filename_BCF_Text, string Filename_Genotype_Matrix, string Filename_Accession_Atlas, string Filename_Marker_Atlas, int Shif_Col, int Min_AF, int Max_Unknow, int Max_Hetero, int Max_Length_Dif_Allele);
int line_filtering_test(string line, char delim, int Shif_Col, int Min_AF, int Max_Unknow, int Max_Hetero, int Min_Length_Dif_Allele);
double get_AF(string str_AF);
int get_Length_Dif(string str_REF, string str_ALT);
bool is_unknown(string str_genotype);
bool is_hetero(string str_genotype);

int main(int argc, char *argv[])
{
    string Filename_BCF_Text;
    string Filename_Genotype_Matrix;
    string Filename_Accession_Atlas;
    string Filename_Marker_Atlas;
    int Shif_Col;
    int Min_AF;
    int Max_Unknow;
    int Max_Hetero;
    int Max_Length_Dif_Allele;
    string sLine;

    if(1==parse_cmd_line(argc, argv, Filename_BCF_Text, Filename_Genotype_Matrix, Filename_Accession_Atlas, Filename_Marker_Atlas, Shif_Col, Min_AF, Max_Unknow, Max_Hetero, Max_Length_Dif_Allele))
    return 1;

    if (1== GMEF(Filename_BCF_Text, Filename_Genotype_Matrix, Filename_Accession_Atlas, Filename_Marker_Atlas, Shif_Col, Min_AF, Max_Unknow, Max_Hetero, Max_Length_Dif_Allele))
    return 1;

    cout << "Genotype Marker Extracting and Filtering Finsihed!" << endl;
    return 0;
}

int GMEF(string Filename_BCF_Text, string Filename_Genotype_Matrix, string Filename_Accession_Atlas, string Filename_Marker_Atlas, int Shif_Col, int Min_AF, int Max_Unknow, int Max_Hetero, int Min_Length_Dif_Allele)
{

    ifstream BCF_Text_File;
	ofstream Genotype_Matrix_File;
	ofstream Marker_Atlas_File;
	ofstream Accession_Atlas_File;

	string sLine;
	long Row_Count   =0;
	long Col_Count =0;

	BCF_Text_File.open(Filename_BCF_Text.c_str(), ios::in);
	Genotype_Matrix_File.open(Filename_Genotype_Matrix.c_str(), ios::out);
	Marker_Atlas_File.open(Filename_Marker_Atlas.c_str(), ios::out);
	Accession_Atlas_File.open(Filename_Accession_Atlas.c_str(), ios::out);

    while(getline(BCF_Text_File, sLine))
	{
		long New_Count =0;
		cout<<"Row_Count=" << Row_Count<<endl;
		if(sLine.empty()) ; // Ignore empty lines
		else
		{
		    stringstream ss(sLine);
			vector <string> s_v;
			string item;
			char delim ='\t';
			std::size_t found = sLine.find(delim);
			if(found ==std::string::npos) delim=',';
            if(Row_Count ==0) // means the first head line
            {
                 while(getline(ss, item, delim))
                 {
                    if(New_Count < Shif_Col) Marker_Atlas_File << item;
                    if(New_Count < Shif_Col-1) Marker_Atlas_File << delim;
                    if(New_Count == Shif_Col)  Accession_Atlas_File<<item;
                    if(New_Count > Shif_Col)
                    {
                       Accession_Atlas_File<<delim;
                       Accession_Atlas_File<<item;
                    }
                    New_Count++;
                 }
                 Marker_Atlas_File <<endl;
                 Accession_Atlas_File<<endl;
                 Col_Count = New_Count;
            }
            else  //means the real data line
            {
                if(0==line_filtering_test(sLine, delim, Shif_Col, Min_AF, Max_Unknow, Max_Hetero, Min_Length_Dif_Allele))
                {
                    while(getline(ss, item, delim))
                    {
                       if(New_Count < Shif_Col) Marker_Atlas_File << item;
                       if(New_Count < Shif_Col-1) Marker_Atlas_File << delim;
                       if(New_Count == Shif_Col)  Genotype_Matrix_File<<item;
                       if(New_Count > Shif_Col)
                       {
                          Genotype_Matrix_File<<delim;
                          Genotype_Matrix_File<<item;
                       }
                       New_Count++;
                    }

                    Marker_Atlas_File <<endl;
                    Genotype_Matrix_File <<endl;

                    if(New_Count != Col_Count)
                    {
                       cerr<<"The inputting Data Format is not correct!"<<endl;
                       return 1;
                    }
                }
            }
			Row_Count++;

		}
    }

    Genotype_Matrix_File.close();
	Marker_Atlas_File.close();
	Accession_Atlas_File.close();
    return 0;
}

// This function is used todouble get_AF(string str_AF)
int line_filtering_test(string line, char delim, int Shif_Col, int Min_AF, int Max_Unknow, int Max_Hetero, int Min_Length_Dif_Allele)
{
     stringstream ss(line);
     vector <string> s_v;
     string item;
     string Current_REF;
     string Current_ALT;
     float Current_AF;
     int Current_Length_Dif;
     int Count_Hetero =0;
     int Count_Unknown=0;
     int New_Count =0;
     while(getline(ss, item, delim))
     {
        if(New_Count == 2)  Current_REF =item;
        if(New_Count == 3)  Current_ALT =item;
        if(New_Count == 6)  Current_AF  =get_AF(item);

        if(New_Count >= Shif_Col) // come to genotype
        {
           if(is_unknown(item)) Count_Unknown++;
           if(is_hetero(item))  Count_Hetero++;
        }
        New_Count++;
     }

     Current_Length_Dif =get_Length_Dif(Current_REF,Current_ALT);

 //    cout <<"Current_AF=" << Current_AF*100 << "Current_Length_Dif=" << Current_Length_Dif << "Count_Unknown=" << Count_Unknown <<"Count_Hetero=" << Count_Hetero <<endl;
     if(Current_AF*100< Min_AF) return 1;
     if(Current_Length_Dif <Min_Length_Dif_Allele) return 1;
     if(Count_Unknown*100.0/(New_Count-Shif_Col) >Max_Unknow) return 1;
     if(Count_Hetero*100.0/(New_Count-Shif_Col) >Max_Hetero) return 1;

     return 0;
}

double get_AF(string str_AF)
{
   double  Ret_AF=1.0;
   stringstream AF_stream(str_AF);
   string af;
   while(getline(AF_stream,af,','))
   {
      double value = atof(af.c_str());
      if(Ret_AF>value) Ret_AF=value;
      if(Ret_AF>1-value) Ret_AF=1-value;
   }

   return Ret_AF;
}

bool is_unknown(string str_genotype)
{
    if(str_genotype.compare("./.")==0)
       return true;
    else
       return false;
}

bool is_hetero(string str_genotype)
{
    char genotype1=atol(str_genotype.substr(0,1).c_str());
    char genotype2=atol(str_genotype.substr(2,1).c_str());
    if(genotype1!=genotype2)
      return true;
    else
      return false;
}


int get_Length_Dif(string str_REF, string str_ALT)
{
   vector <int> length_vec;
   length_vec.push_back(str_REF.length());
   stringstream alt_stream(str_ALT);
   string alt_genotype;
   while(getline(alt_stream,alt_genotype,','))
   {
      length_vec.push_back(alt_genotype.length());
   }
   sort(length_vec.begin(),length_vec.end());

   int lenth_prev= length_vec.back();
   int lenth_dif = lenth_prev;
   length_vec.pop_back();
   while(!length_vec.empty())
   {
      int lenth_var = lenth_prev-length_vec.back();
      if(lenth_var< lenth_dif) lenth_dif= lenth_var;
      lenth_prev= length_vec.back();
      length_vec.pop_back();
   }
   return lenth_dif;
}

int parse_cmd_line(int argc, char* argv[], string & Filename_BCF_Text, string & Filename_Genotype_Matrix, string & Filename_Accession_Atlas, string & Filename_Marker_Atlas, int & Shif_Col, int & Min_AF, int & Max_Unknow, int & Max_Hetero, int & Max_Length_Dif_Allele)
{
    if((argc!=2)&&(argc <19))
	{
		cerr <<"Please specific the correct parameters for !" <<endl;
		return 1;
	}
	else
	{
	   for (int i=1; i<argc;i++)
       {
	       string arg=argv[i];

           if((arg=="-b")||(arg=="-B"))
	       {
              if(i+1<argc)
		      {
		         Filename_BCF_Text = argv[i+1];
		      }
		      else
		      {
			     cerr<<"Parse cmd_line fail, Need to clearly specific the filename of inputting BCF_Text !" <<endl;
			     return 1; // Parse cmd_line fail
		      }
		    }

		    if((arg=="-g")||(arg=="-G"))
	        {
              if(i+1<argc)
		      {
		         Filename_Genotype_Matrix = argv[i+1];
		      }
		      else
		      {
			     cerr<<"Parse cmd_line fail, Need to clearly specific the filename of outputing Genotype Matrix Data!" <<endl;
			     return 1; // Parse cmd_line fail
		      }
		    }

		    if((arg=="-a")||(arg=="-A"))
	        {
              if(i+1<argc)
		      {
		         Filename_Accession_Atlas = argv[i+1];
		      }
		      else
		      {
			     cerr<<"Parse cmd_line fail, Need to clearly specific the filename of outputing Accession Atlas!" <<endl;
			     return 1; // Parse cmd_line fail
		      }
		    }

		    if((arg=="-m")||(arg=="-M"))
	        {
              if(i+1<argc)
		      {
		         Filename_Marker_Atlas = argv[i+1];
		      }
		      else
		      {
			     cerr<<"Parse cmd_line fail, Need to clearly specific the filename of outputing Marker Atlas!!" <<endl;
			     return 1; // Parse cmd_line fail
		      }
		    }

	       if((arg=="-s")||(arg=="-S"))
	       {
              if(i+1<argc)
		      {
		         Shif_Col =atoi(argv[i+1]);
		      }
		      else
		      {
			     cerr<<"Parse cmd_line fail, Need to clearly specific the Shift Column number for the real genotype data!" <<endl;
			     return 1; // Parse cmd_line fail
		      }
		    }

           if((arg=="-f")||(arg=="-F"))
	       {
              if(i+1<argc)
		      {
		         Min_AF =atoi(argv[i+1]);
		      }
		      else
		      {
			     cerr<<"Parse cmd_line fail, Need to clearly specific the Min_AF!" <<endl;
			     return 1; // Parse cmd_line fail
		      }
		    }


	       if((arg=="-u")||(arg=="-U"))
	       {
              if(i+1<argc)
		      {
		         Max_Unknow =atoi(argv[i+1]);
		      }
		      else
		      {
			     cerr<<"Parse cmd_line fail, Need to clearly specific the Maximum Unknow Genotype Number!" <<endl;
			     return 1; // Parse cmd_line fail
		      }
		    }


           if((arg=="-e")||(arg=="-E"))
	       {
              if(i+1<argc)
		      {
		         Max_Hetero =atoi(argv[i+1]);
		      }
		      else
		      {
			     cerr<<"Parse cmd_line fail, Need to clearly specific the Maximum Hetero Genotype Number!" <<endl;
			     return 1; // Parse cmd_line fail
		      }
		    }


           if((arg=="-d")||(arg=="-D"))
	       {
              if(i+1<argc)
		      {
		         Max_Length_Dif_Allele =atoi(argv[i+1]);
		      }
		      else
		      {
			     cerr<<"Parse cmd_line fail, Need to clearly specific the Maximum Length Difference among a allele!" <<endl;
			     return 1; // Parse cmd_line fail
		      }
		    }

            if((arg=="-h")||(arg=="-H"))
	        {
			     cout<<"Welcome to use this program to get the filtered genetic variant filetering" <<endl;
				 cout << "The usuage of input parameter arguments are listed as followings:" <<endl;
				 cout << "-h or -H: Output this help usuage message" <<endl;
				 cout << "-b or -B:  The filename of inputting BCF_Text " <<endl;
				 cout << "-g or -G:  The filename of outputing Genotype Matrix Data " <<endl;
				 cout << "-a or -A:  The filename of outputing Accession Atlas" <<endl;
				 cout << "-m or -M:  The filename of outputing Marker Atlas " <<endl;
				 cout << "-s or -S: The Shift Column number for the real genotype data" <<endl;
				 cout << "-f or -F: The the allowed Min_AF(Mininmum Allele Frequency)" <<endl;
				 cout << "-u or -U: The allowed maximum unknown genetype number" <<endl;
				 cout << "-e or -E: The allowed maximum heterogenous genetype number" <<endl;
				 cout << "-d or -D: The length difference among different alleles"<<endl;
			     return 1; // Parse cmd_line fail
		    }
         }
     }
     return 0;
}
